#include <string.h>
#include "Spriteset.hpp"
#include "packed.hpp"
#include "x++.hpp"



////////////////////////////////////////////////////////////////////////////////

sSpriteset::sSpriteset()
: m_FrameWidth(0)
, m_FrameHeight(0)

, m_BaseX1(0)
, m_BaseY1(0)
, m_BaseX2(0)
, m_BaseY2(0)
{
}

////////////////////////////////////////////////////////////////////////////////

sSpriteset::~sSpriteset()
{
  Destroy();
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::Create(int num_directions, int num_frames)
{
  Destroy();

  m_FrameWidth = 16;
  m_FrameHeight = 32;

  m_Directions.resize(num_directions);
  for (int i = 0; i < num_directions; i++) {
    m_Directions[i].frames.resize(num_frames);
    for (int j = 0; j < num_frames; j++) {
      sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
      sprite->SetDelay(8);
      m_Directions[i].frames[j] = sprite;
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

#define STRUCT_NAME SPRITESET_HEADER
#include "begin_packed_struct.h"
  byte signature[4];
  word version;
  word num_frames;
  word frame_width;
  word frame_height;
  word num_directions;
  word base_x1;
  word base_y1;
  word base_x2;
  word base_y2;
  byte reserved[106];
#include "end_packed_struct.h"

#define STRUCT_NAME DIRECTION_HEADER
#include "begin_packed_struct.h"
  word num_frames;
  byte reserved[62];
#include "end_packed_struct.h"

#define STRUCT_NAME FRAME_HEADER
#include "begin_packed_struct.h"
  word width;
  word height;
  word delay;
  byte reserved[26];
#include "end_packed_struct.h"

ASSERT_STRUCT_SIZE(SPRITESET_HEADER, 128)
ASSERT_STRUCT_SIZE(DIRECTION_HEADER, 64)
ASSERT_STRUCT_SIZE(FRAME_HEADER,     32)

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Load(const char* filename)
{
  // open file
  FILE* file = fopen(filename, "rb");
  bool result = LoadFromFile(file);
  fclose(file);

  return result;
} 

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::LoadFromFile(FILE* file)
{
  // open file
  if (file == NULL)
    return false;

  // read the header
  SPRITESET_HEADER header;
  fread(&header, 1, sizeof(header), file);

  // validate header
  if (memcmp(header.signature, ".rss", 4) != 0 ||
      (header.version != 1 && header.version != 2))
  {
    return false;
  }

  Destroy();

  m_FrameWidth  = header.frame_width;
  m_FrameHeight = header.frame_height;

  m_BaseX1 = header.base_x1;
  m_BaseY1 = header.base_y1;
  m_BaseX2 = header.base_x2;
  m_BaseY2 = header.base_y2;

  if (header.version == 1)  // VERSION ONE
  {
    m_Directions.resize(8);
    for (int i = 0; i < 8; i++)
    {
      m_Directions[i].frames.resize(8);
      for (int j = 0; j < 8; j++)
      {
        sSprite* sprite = new sSprite(header.frame_width, header.frame_height);
        fread(sprite->GetPixels(), sizeof(RGBA), header.frame_width * header.frame_height, file);
        sprite->SetDelay(8);

        m_Directions[i].frames[j] = sprite;
      }
    }
  }
  else                      // VERSION TWO
  {
    m_Directions.resize(header.num_directions);

    for (int i = 0; i < header.num_directions; i++)
    {
      // read the direction header
      DIRECTION_HEADER direction_header;
      fread(&direction_header, 1, sizeof(direction_header), file);

      m_Directions[i].frames.resize(direction_header.num_frames);

      for (int j = 0; j < direction_header.num_frames; j++)
      {
        // read the frame header
        FRAME_HEADER frame_header;
        fread(&frame_header, 1, sizeof(frame_header), file);

        // some backwards compatibility hacking
        if (m_FrameWidth == 0 || m_FrameHeight == 0)
        {
          m_FrameWidth  = frame_header.width;
          m_FrameHeight = frame_header.height;
        }

        sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
        sprite->SetDelay(frame_header.delay);
        fread(sprite->GetPixels(), sizeof(RGBA), m_FrameWidth * m_FrameHeight, file);

        m_Directions[i].frames[j] = sprite;
      }
    }
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Save(const char* filename) const
{
  // open file
  FILE* file = fopen(filename, "wb");
  if (file == NULL)
    return false;

  // fill header
  SPRITESET_HEADER header;
  memset(&header, 0, sizeof(header));
  memcpy(header.signature, ".rss", 4);
  header.version        = 2;
  header.frame_width    = m_FrameWidth;
  header.frame_height   = m_FrameHeight;
  header.num_directions = m_Directions.size();
  header.base_x1        = m_BaseX1;
  header.base_y1        = m_BaseY1;
  header.base_x2        = m_BaseX2;
  header.base_y2        = m_BaseY2;

  // write header
  fwrite(&header, 1, sizeof(header), file);

  // write all of the directions
  for (int i = 0; i < m_Directions.size(); i++)
  {
    // fill direction header
    DIRECTION_HEADER direction_header;
    memset(&direction_header, 0, sizeof(direction_header));
    direction_header.num_frames = m_Directions[i].frames.size();

    // write it
    fwrite(&direction_header, 1, sizeof(direction_header), file);

    // write all of the frames
    for (int j = 0; j < m_Directions[i].frames.size(); j++)
    {
      // fill frame header
      FRAME_HEADER frame_header;
      memset(&frame_header, 0, sizeof(frame_header));
      frame_header.delay  = m_Directions[i].frames[j]->GetDelay();

      // write it
      fwrite(&frame_header, 1, sizeof(frame_header), file);

      // write the frame
      fwrite(m_Directions[i].frames[j]->GetPixels(), sizeof(RGBA), m_FrameWidth * m_FrameHeight, file);
    }
  }
  
  fclose(file);
  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Import_BMP(const char* filename, int frame_width, int frame_height, RGBA transparent)
{
  CImage32 image;
  if (!image.Load(filename))
    return false;

  if (image.GetWidth() % frame_width || image.GetHeight() % frame_height)
    return false;

  Destroy();

  // replace for the transparent color
  for (int y = 0; y < image.GetHeight(); y++)
    for (int x = 0; x < image.GetWidth(); x++)
      if (image.GetPixels()[y * image.GetWidth() + x].red   == transparent.red &&
          image.GetPixels()[y * image.GetWidth() + x].green == transparent.green &&
          image.GetPixels()[y * image.GetWidth() + x].blue  == transparent.blue)
      {
        image.GetPixels()[y * image.GetWidth() + x].alpha = 0;
      }

  m_FrameWidth = frame_width;
  m_FrameHeight = frame_height;
  int numRows = image.GetHeight() / frame_height;
  int numFrames = image.GetWidth() / frame_width;

  m_Directions.resize(numRows);
  for (int i = 0; i < numRows; i++)
  {
    m_Directions[i].frames.resize(numFrames);

    for (int j = 0; j < numFrames; j++)
    {
      sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
      sprite->SetDelay(8);
      memset(sprite->GetPixels(), 0, m_FrameWidth * m_FrameHeight * sizeof(RGBA));

      m_Directions[i].frames[j] = sprite;
    }
  }

  // grabs the data in a gigantic loops 'o doom...
  for (int j=0; j<numRows; j++)
    for (int i=0; i<numFrames; i++)
      for (int k=0; k<m_FrameHeight; k++)
        memcpy(m_Directions[j].frames[i]->GetPixels() + (k*m_FrameWidth),
               image.GetPixels() + (j*m_FrameHeight*image.GetWidth()) + (i*m_FrameWidth) + (k*image.GetWidth()),
               m_FrameWidth*sizeof(RGBA));

  return true;
}

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Import_CHR(const char* filename, const char* palette_file)
{
  // read the palette
  RGB palette[256];
  FILE* file = fopen(palette_file, "rb");
  if (file == NULL)
    return false;
  fread(palette, 256, sizeof(RGB), file);
  fclose(file);

  // determine which type of .chr it is
  file = fopen(filename, "rb");
  if (file == NULL)
    return false;

  // get the version of the .chr
  byte version;
  version = (byte)fgetc(file);
  fseek(file, 0, SEEK_SET);

  // import the .chr
  bool success;
  if (version == 0)
    success = Import_CHR1(file, palette);
  else
    success = Import_CHR2(file, palette);

  fclose(file);
  return success;
}

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Export_PNG(const char* filename)
{
  int i,j,k;
  int maxRowFrames = 0;

  // find the maximum rows frames to establish the width of the image
  for (j=0; j<m_Directions.size(); j++)
    if (GetNumFrames(j) > maxRowFrames)
      maxRowFrames = GetNumFrames(j);

  CImage32 image(maxRowFrames * m_FrameWidth + (1 * maxRowFrames) + 1,
                 m_Directions.size() * m_FrameHeight + (1 * m_Directions.size()) + 1);

  image.Rectangle(0, 0, 
                  maxRowFrames * m_FrameWidth + (1 * maxRowFrames),
                  m_Directions.size() * m_FrameHeight + (1 * m_Directions.size()),
                  rgbaWhite);

  // grabs the data in a gigantic loops 'o doom...
  for (j=0; j<m_Directions.size(); j++)
    for (i=0; i<GetNumFrames(j); i++)
      for (k=0; k<m_FrameHeight; k++)
        memcpy(image.GetPixels() + (((j*m_FrameHeight) + 1) * image.GetWidth()) + (i*m_FrameWidth) + (i*1) + (j*image.GetWidth()) + (k*image.GetWidth()) + 1,
               m_Directions[j].frames[i]->GetPixels() + (k*m_FrameWidth),
               m_FrameWidth*sizeof(RGBA));

  return image.Save(filename);
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetFrameWidth() const
{
  return m_FrameWidth;
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetFrameHeight() const
{
  return m_FrameHeight;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::ResizeFrames(int width, int height)
{
  for (int i = 0; i < m_Directions.size(); i++) {
    for (int j = 0; j < m_Directions[i].frames.size(); j++) {
      m_Directions[i].frames[j]->Resize(width, height);
    }
  }

  m_FrameWidth  = width;
  m_FrameHeight = height;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::InsertDirection(int direction)
{
  Direction new_direction;

  new_direction.frames.resize(1);
  sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
  sprite->SetDelay(8);
  new_direction.frames[0] = sprite;

  m_Directions.insert(m_Directions.begin() + direction, new_direction);
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::DeleteDirection(int direction)
{
  Direction& d = m_Directions[direction];
  for (int i = 0; i < d.frames.size(); i++) {
    delete d.frames[i];
  }

  m_Directions.erase(m_Directions.begin() + direction);
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::AppendDirection()
{
  // make an empty sprite
  sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
  sprite->SetDelay(8);

  // add an empty direction
  Direction new_direction;
  new_direction.frames.resize(1);
  new_direction.frames[0] = sprite;
  m_Directions.insert(m_Directions.end(), new_direction);
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::InsertFrame(int direction, int frame)
{
  // new frame
  sSprite* new_frame = new sSprite(m_FrameWidth, m_FrameHeight);
  new_frame->SetDelay(8);

  // direction
  Direction& d = m_Directions[direction];
  d.frames.insert(d.frames.begin() + frame, new_frame);
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::DeleteFrame(int direction, int frame)
{
  // direction
  Direction& d = m_Directions[direction];
  delete d.frames[frame];
  d.frames.erase(d.frames.begin() + frame);
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::AppendFrame(int direction)
{
  // new frame
  sSprite* new_frame = new sSprite(m_FrameWidth, m_FrameHeight);
  new_frame->SetDelay(8);

  // direction
  Direction& d = m_Directions[direction];
  d.frames.insert(d.frames.end(), new_frame);
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetNumDirections() const
{
  return m_Directions.size();
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetNumFrames(int direction) const
{
  return m_Directions[direction].frames.size();
}

////////////////////////////////////////////////////////////////////////////////

sSprite&
sSpriteset::GetFrame(int direction, int frame)
{
  return *m_Directions[direction].frames[frame];
}

////////////////////////////////////////////////////////////////////////////////

const sSprite&
sSpriteset::GetFrame(int direction, int frame) const
{
  return *m_Directions[direction].frames[frame];
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetBaseX1() const
{
  return m_BaseX1;
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetBaseY1() const
{
  return m_BaseY1;
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetBaseX2() const
{
  return m_BaseX2;
}

////////////////////////////////////////////////////////////////////////////////

int
sSpriteset::GetBaseY2() const
{
  return m_BaseY2;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::SetBaseX1(int x)
{
  m_BaseX1 = x;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::SetBaseY1(int x)
{
  m_BaseY1 = x;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::SetBaseX2(int x)
{
  m_BaseX2 = x;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::SetBaseY2(int x)
{
  m_BaseY2 = x;
}

////////////////////////////////////////////////////////////////////////////////

void
sSpriteset::Destroy()
{
  for (int i = 0; i < m_Directions.size(); i++) {
    for (int j = 0; j < m_Directions[i].frames.size(); j++) {
      delete m_Directions[i].frames[j];
    }
    m_Directions[i].frames.resize(0);
  }
  m_Directions.resize(0);
}

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Import_CHR1(FILE* file, RGB palette[256])
{
  // create the frames...
  m_FrameWidth = 16;
  m_FrameHeight = 32;

  m_Directions.resize(9);
  for (int i = 0; i < 9; i++)
  {
    m_Directions[i].frames.resize(i == 8 ? 10 : 8);
    int k = (i == 8 ? 2 : 0);

    for (int j = 0; j < (8 + k); j++)
    {
      sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
      sprite->SetDelay(8);

      m_Directions[i].frames[j] = sprite;
    }
  }

  // allocate .chr frames and read them
  RGBA chr_frames[30][16 * 32];

  for (int i = 0; i < 30; i++)
  {
    for (int y = 0; y < 32; y++)
      for (int x = 0; x < 16; x++)
      {
        byte c = (byte)fgetc(file);

        RGBA* p = chr_frames[i] + (y * 16 + x);
        p->red   = palette[c].red   * 4;
        p->green = palette[c].green * 4;
        p->blue  = palette[c].blue  * 4;
        p->alpha = (byte)(c == 0 ? 0 : 255);
      }
  }

#define CopyFrame(direction, row, frame, column)             \
{                                                            \
  memcpy(m_Directions[direction].frames[frame]->GetPixels(), \
         chr_frames[row * 5 + column],                       \
         16 * 32 * sizeof(RGBA));                            \
}

#define CopyRow(direction, row)    \
{                                  \
  CopyFrame(direction, row, 0, 0); \
  CopyFrame(direction, row, 1, 1); \
  CopyFrame(direction, row, 2, 2); \
  CopyFrame(direction, row, 3, 1); \
  CopyFrame(direction, row, 4, 0); \
  CopyFrame(direction, row, 5, 3); \
  CopyFrame(direction, row, 6, 4); \
  CopyFrame(direction, row, 7, 3); \
}

  // copy frames to spriteset
  CopyRow(0, 1);
  CopyRow(1, 2);
  CopyRow(2, 2);
  CopyRow(3, 2);
  CopyRow(4, 0);
  CopyRow(5, 3);
  CopyRow(6, 3);
  CopyRow(7, 3);
  for (int i=0; i<10; i++)
    CopyFrame(8, 4, i, i);

  return true;
}

////////////////////////////////////////////////////////////////////////////////

#define STRUCT_NAME CHR2_HEADER
#include "begin_packed_struct.h"
  byte version;
  word width;
  word height;
  word hotspot_left;
  word hotspot_top;
  word hotspot_width;
  word hotspot_height;
  word num_frames;
#include "end_packed_struct.h"

////////////////////////////////////////////////////////////////////////////////

bool
sSpriteset::Import_CHR2(FILE* file, RGB palette[256])
{
  CHR2_HEADER header;
  fread(&header, sizeof(header), 1, file);

   // because we want to get the extra frames in the set, so I'm building
  // a table, and then parsing it. This is so that I'm marking down the 
  // frames that's used, and the unused shall be in a seperate direction.
  // Finally we then do something about it.
  bool* frameUsed;
  bool  extraFramesUsed;
  RGBA* frames;
  long  dataLength;
  dword walkcodeLength[4];
  char* walkcode[4];
  int*  walkcodeValue[4];
  int* walkcodeType[4];
  int   walkcodeNumArg[4];
  int frameCount[5];
  int i,j,k,l;
  
  if (header.version != 2)
  return false;

  Destroy();

  //grab all the data then expand it.
  frameCount[0] = frameCount[1] = frameCount[2] = frameCount[3] = 1;
  frameCount[4] = 0;

  frameUsed = new bool[header.num_frames];
  memset(frameUsed, 0, header.num_frames);
  frames = new RGBA[header.num_frames * header.width * header.height];
  j=0;
    
  fread(&dataLength, 1, 4, file);
  for (i=0; i<dataLength; i++)
  {
    byte value = getc(file);
    
    if (value<255)
    {
      frames[j].red = palette[value].red * 4;
      frames[j].blue = palette[value].blue * 4;
      frames[j].green = palette[value].green * 4;
      frames[j].alpha = (byte)(value == 0 ? 0 : 255);
      j++;
    }
    else if (value == 255)
    {
      byte runlength = getc(file);
      value = getc(file);
      for(k=0; k<runlength; k++)
      {
        frames[j].red = palette[value].red * 4;
        frames[j].blue = palette[value].blue * 4;
        frames[j].green = palette[value].green * 4;
        frames[j].alpha = (byte)(value == 0 ? 0 : 255);
        j++;
      }
      i+=2;
    }
  }

  for (i=0; i<4; i++)
  {
    walkcodeType[i] = new int[1];
    walkcodeValue[i] = new int[1];
    walkcodeType[i][0] = true;
	  fread(&walkcodeValue[i][0], 1, sizeof(long), file);
  }

  // read the movement script
  for (i=0; i<4; i++)
  {
    fread(&walkcodeLength[i], 1, sizeof(dword), file);
    walkcode[i] = new char[walkcodeLength[i]];
    fread(walkcode[i], 1, walkcodeLength[i], file);
  }

  // translate the movement script
  for (i=0; i<4; i++)
  {
    k=1;
    for (j=0; j<(int)walkcodeLength[i];)
      if (walkcode[i][j] == 'F' || walkcode[i][j] == 'W')
      {
        char command;
        char token[10];
        int* tempwalkcodeType;
        int* tempwalkcodeValue;
        command = walkcode[i][j];
        l=0;
        j++;

        while (j < (int)walkcodeLength[i] && walkcode[i][j] == ' ') j++;
        while (j < (int)walkcodeLength[i] && walkcode[i][j] >= 48 &&
               walkcode[i][j] <= 57)
        {
          token[l] = walkcode[i][j];
          l++;
          j++;
        }
        
        token[l] = 0;

        tempwalkcodeType = new int[k + 1];
        tempwalkcodeValue = new int[k + 1];
        memcpy(tempwalkcodeType, walkcodeType[i], k*sizeof(int));
        memcpy(tempwalkcodeValue, walkcodeValue[i], k*sizeof(int));
        tempwalkcodeType[k] = (command == 'F' ? 1 : 0);
        tempwalkcodeValue[k] = atoi(token);
        delete[] walkcodeType[i]; delete[] walkcodeValue[i];
        walkcodeType[i] = tempwalkcodeType;
        walkcodeValue[i] = tempwalkcodeValue;

        if (command == 'F')
        {
        frameUsed[atoi(token)] = true;
        frameCount[i]++;
        }
        k++;
      }
      else
        j++;

    walkcodeNumArg[i] = k;
  }

  // check for extra frames
  for (i=0; i<header.num_frames; i++)
    if (!frameUsed[i])
      frameCount[4]++;
  if (frameCount[4] > 0)
    extraFramesUsed = true;
  else
    extraFramesUsed = false;


  // okay, now build the data the painful way (Create() can't handle this)
  // and transfer ALL the data.
  m_FrameWidth = header.width;
  m_FrameHeight = header.height;

  m_Directions.resize(8 + extraFramesUsed);
  for (int i = 0; i < 8 + extraFramesUsed; i++)
  {
    switch(i)
    {
      case EAST: 
      case SOUTHEAST: 
      case NORTHEAST:
        l = 0;
        break;

      case WEST:
      case NORTHWEST:
      case SOUTHWEST:
        l = 1;
        break;

      case NORTH: l = 2; break;
      case SOUTH: l = 3; break;
      case 8: l = 4; break;
    }

    m_Directions[i].frames.resize(l);

    for (int j = 0; j < frameCount[l]; j++)
    {
      sSprite* sprite = new sSprite(m_FrameWidth, m_FrameHeight);
      sprite->SetDelay(0);

      m_Directions[i].frames[j] = sprite;
    }

  }

  // left right up down special
#define CopySide(direction, theside)                                    \
{                                                                       \
  k=-1;                                                                 \
  for (i=0; i<walkcodeNumArg[theside]; i++)                             \
    if (walkcodeType[theside][i] == 1)                                  \
    {                                                                   \
      k++;                                                              \
      if (k < frameCount[theside])                                      \
      memcpy(m_Directions[direction].frames[k]->GetPixels(),             \
             frames + (walkcodeValue[theside][i]*header.width*header.height),\
             header.width * header.height * 4);                         \
    }                                                                   \
    else                                                                \
    {                                                                   \
      m_Directions[direction].frames[k]->SetDelay(m_Directions[direction].frames[k]->GetDelay() + walkcodeValue[theside][i]);\
    }                                                                   \
}

  CopySide(EAST, 1);
  CopySide(NORTHEAST, 1);
  CopySide(SOUTHEAST, 1);
  CopySide(WEST, 0);
  CopySide(NORTHWEST, 0);
  CopySide(SOUTHWEST, 0);
  CopySide(NORTH, 2);
  CopySide(SOUTH, 3);

  j = 0;
  if (extraFramesUsed)
    {
    for (i=0; i<header.num_frames; i++)
      if (!frameUsed[i])
      {
        memcpy(m_Directions[8].frames[j]->GetPixels(),
               frames + (i * header.width * header.height),
               header.width * header.height * 4);
        j++;
      }
    }

  delete[] frameUsed;
  delete[] frames;
  for (i=0; i<4; i++)
    {
    delete[] walkcode[i];
    delete[] walkcodeValue[i];
    delete[] walkcodeType[i];
    }

  return true;

}


////////////////////////////////////////////////////////////////////////////////
