#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "x++.hpp"
#include "configfile.h"

#include "engine.hpp"
#include "engineaudio.hpp"
#include "render.h"
#include "rendersort.hpp"
#include "inputx.h"
#include "animation.h"
#include "sfont.hpp"
#include "simage.hpp"

#include "input.h"
#include "timer.h"
#include "filesystem.h"


const int FPS     = 60;
const int MIN_FPS = 5;


////////////////////////////////////////////////////////////////////////////////

CGameEngine::CGameEngine(const SSystemFiles* system_files)
: m_SystemFiles(system_files)

, ShouldExitEngine(false)
, ExitEngineMessage("")

, GameScript(NULL)

, MapLoaded(false)
, MapScript(NULL)
, OnTrigger(false)

, CurrentDoodad(-1)
{
  ME_InitColorMask();
}

////////////////////////////////////////////////////////////////////////////////

CGameEngine::~CGameEngine()
{
  ME_RemovePersons();
  ME_RemoveDoodads();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::Run(const char* game)
{
  EnterDirectory(game);

  InitializeScriptSystem(this);

  std::string gamepackage;
  gamepackage = game;
  gamepackage += ".spk";
  m_FileServer.LoadSPK(gamepackage.c_str());

  // run the game engine
  if (InitializeGame())
  {
    if (!ShouldExitEngine)
      RunGame();
    ShutdownGame();
  }

  ShutdownScriptSystem();

  LeaveDirectory();
  ShowErrorScreen();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::Exit()
{
  ExitWithMessage("");
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ExitWithMessage(const char* message)
{
  ExitEngineMessage = message;
  ShouldExitEngine = true;
}

////////////////////////////////////////////////////////////////////////////////

bool
CGameEngine::ShouldExit()
{
  return ShouldExitEngine;
}

////////////////////////////////////////////////////////////////////////////////

IMAGE
CGameEngine::LoadImage(const char* filename)
{
  char path[1024];
  sprintf(path, "images/%s", filename);
  FILE* file = m_FileServer.GetFile(path);

  IMAGE image = SLoadImage(file);
  if (image == NULL)
  {
    char message[1024];
    sprintf(message, "Could not load image '%s'", filename);
    ExitWithMessage(message);
  }
  return image;
}

////////////////////////////////////////////////////////////////////////////////

CImage32*
CGameEngine::LoadSurface(const char* filename)
{
  char path[1024];
  CImage32* image = new CImage32();
  
  sprintf(path, "images/%s", filename);
  FILE* file = m_FileServer.GetFile(path);

  if (!image->LoadFromFile(file))
  {
    fclose(file);

    char message[1024];
    sprintf(message, "Could not load image '%s'", filename);
    ExitWithMessage(message);
  }

  fclose(file);
  return image;
}
////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetFont(const char* filename)
{
  char path[1024];
  
  sprintf(path, "fonts/%s", filename);
  FILE* file = m_FileServer.GetFile(path);

  if (!Font.Load(file))
  {
    char error_message[1024];
    sprintf(error_message, "Error: Could not load font '%s'", filename);
    ExitWithMessage(error_message);
  }
}

////////////////////////////////////////////////////////////////////////////////

SFONT*
CGameEngine::GetFont()
{
  return &Font;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetWindowStyle(const char* filename)
{
  char path[1024];
  sprintf(path, "windowstyles/%s", filename);
  FILE* file = m_FileServer.GetFile(path);

  if (!WindowStyle.Load(file))
  {
    char error_message[1024];
    sprintf(error_message, "Error: Could not load window style '%s'", filename);
    ExitWithMessage(error_message);
  }
}

////////////////////////////////////////////////////////////////////////////////

SWINDOWSTYLE*
CGameEngine::GetWindowStyle()
{
  return &WindowStyle;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::AddFrameHook(int update_rate, const char* function)
{
  FrameHook fh;
  fh.current = 0;
  fh.update_rate = update_rate;
  fh.function = function;
  FrameHooks.push_back(fh);
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::RemoveFrameHook(const char* function)
{
  for (int i = 0; i < FrameHooks.size(); i++)
    if (FrameHooks[i].function == function)
    {
      FrameHooks.erase(FrameHooks.begin() + i);
      return;
    }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::MapEngine()
{
  if (MapLoaded == false)
  {
    ExitWithMessage("MapEngine() called when no map is loaded.  Use ChangeMap() first.");
    return;
  }

  // rendering throttle/meter
  ShowFPS = false;
  FPSKeyPressed = false;
  dword frames_rendered = 0;
  dword last_fps = 0;
  dword last_update = 0;

  dword frames_skipped = 0;

  dword idealtime = GetTime() * FPS;

  ShouldExitEngine = false;
  while (!ShouldExitEngine)
  {
    dword actualtime = GetTime() * FPS;
    idealtime += 1000; // number of milliseconds in a second

    // RENDER PHASE
    if (ShowFPS)
    {
      dword old_frames_rendered = frames_rendered;

      while (actualtime < idealtime || frames_skipped >= MIN_FPS)
      {
        ME_RenderScreen();
        frames_skipped = 0;

        // draw FPS
        char fps_string[80];
        sprintf(fps_string, "%d/%d", last_fps, FPS);
        Font.SetMask(rgbaBlack);
        Font.DrawText(1, 1, fps_string);
        Font.SetMask(rgbaWhite);
        Font.DrawText(0, 0, fps_string);

        FlipScreen();
        frames_rendered++;

        actualtime = GetTime() * FPS;
      }

      if (frames_rendered == old_frames_rendered)
        frames_skipped++;

      if (GetTime() > last_update + 1000)
      {
        last_update = GetTime();
        last_fps = frames_rendered;
        frames_rendered = 0;
      }
    }
    else
    {
      if (actualtime < idealtime || frames_skipped >= MIN_FPS)
      {
        frames_skipped = 0;
        ME_RenderScreen();
        FlipScreen();
        while (actualtime < idealtime)
          actualtime = GetTime() * FPS;
      }
      else
        frames_skipped++;
    }

    // UPDATE PHASE
    UpdateMap();
    ME_ProcessInput();
    ME_ProcessBoundKeys();
  }

  ClearKeyQueue();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::UpdateMap()
{
  if (MapLoaded == false)
  {
    ExitWithMessage("UpdateEngine() called when no map is loaded. Use ChangeMap() first.");
  }

  ME_ProcessFrameHooks();
  ME_ProcessPersons();
  Map.UpdateAnimations();
  ME_UpdateColorMask();
  ME_CheckWarps();
  ME_CheckTriggers();
  ME_CheckDoodads();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::RenderMap()
{
  if (MapLoaded == false)
  {
    ExitWithMessage("RenderMap() called when no map is loaded. Use ChangeMap() first.");
  }

  ME_RenderScreen();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetMap(const char* mapfile)
{
  FILE* file;

  // free the previous map
  if (MapLoaded)
  {
    ME_RemovePersons();
    ME_RemoveDoodads();

    if (MapScript)
      MapScript->ExecuteFunction("LeaveMap");
    delete MapScript;
    MapScript = NULL;

    MapLoaded = false;
  }

  // if mapfile is NULL or "", don't load a new map
  if (mapfile == NULL ||
      strlen(mapfile) == 0)
    return;

  std::string map_path = "maps/";
  map_path += mapfile;
  file = m_FileServer.GetFile(map_path.c_str());

  // load the map
  if (Map.Load(file) == false)
  {
    char exitmessage[520];
    sprintf(exitmessage, "Error: Could not load map '%s'", mapfile);
    ExitWithMessage(exitmessage);
    return;
  }

  // make sure we have a valid base layer
  if (Map.GetMap().GetStartLayer() >= Map.GetMap().GetNumLayers())
    ExitWithMessage("Map has invalid start layer");

  int tile_width  = Map.GetTileset().GetTileWidth();
  int tile_height = Map.GetTileset().GetTileHeight();

  // set coordinates
  Party.location.mapx      = Map.GetMap().GetStartX();
  Party.location.mapy      = Map.GetMap().GetStartY();
  Party.location.maplayer  = Map.GetMap().GetStartLayer();
  Party.location.direction = Map.GetMap().GetStartDirection();

  CameraTracking = true;
  Camera.x       = Party.location.mapx;
  Camera.y       = Party.location.mapy;

  // put all characters at the party coordinates
  for (int i = 0; i < Party.characters.size(); i++)
  {
    Party.characters[i].direction  = Party.location.direction;
    Party.characters[i].mapx       = Party.location.mapx;
    Party.characters[i].mapy       = Party.location.mapy;
    Party.characters[i].maplayer   = Party.location.maplayer;
    Party.characters[i].nextswitch = 0;
    Party.characters[i].walkframe  = 0;
  }

  // put the train at the party coordinates
  for (int i = 0; i < Party.characters.size() * 16; i++)
  {
    Party.train[i].mapx      = Party.location.mapx;
    Party.train[i].mapy      = Party.location.mapy;
    Party.train[i].maplayer  = Party.location.maplayer;
    Party.train[i].direction = Party.location.direction;
  }

  // play music
  std::string music_path;
  music_path = "music/";
  music_path += Map.GetMap().GetMusicFile();
  file = m_FileServer.GetFile(music_path.c_str());
  long fsize = m_FileServer.GetFileLength(music_path.c_str());
  PlayMusic(music_path.c_str(), file, fsize);
  if (ShouldExitEngine)
    return;

  // set up the doodads
  ME_AddPersons();
  ME_AddDoodads();

  // load script
  if (strlen(Map.GetMap().GetScriptFile()))
  {
    char ScriptPath[512];
    sprintf(ScriptPath, "scripts/%s", Map.GetMap().GetScriptFile());
    file = m_FileServer.GetFile(ScriptPath);

    MapScript = new CScript();
    if (!MapScript->Load(ScriptPath, file))
    {
      delete MapScript;
      MapScript = NULL;
      char str[520];
      sprintf(str, "Error: Could not load script '%s' in map '%s'",
              Map.GetMap().GetScriptFile(), mapfile);
      ExitWithMessage(str);
      return;
    }

    // if load flagged an engine error
    if (ShouldExitEngine)
      return;
  }

  if (MapScript)
    MapScript->ExecuteFunction("EnterMap");
  
  MapLoaded = true;
}

////////////////////////////////////////////////////////////////////////////////

SMAP*
CGameEngine::GetMap()
{
  return &Map;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::BindKey(int key, const char* function)
{
  BoundKeys[key] = function;  
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::UnbindKey(int key)
{
  BoundKeys[key] = "";
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetColorMask(RGBA mask, int milliseconds)
{
  PreviousMask    = CurrentMask;
  TargetMask      = mask;
  FadeTime        = milliseconds * 60 / 1000;  // convert from milliseconds to frames
  CurrentFadeTime = 0;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetDoodadFrame(int frame)
{
  if (CurrentDoodad != -1)
    Doodads[CurrentDoodad].current_frame = frame;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetDoodadObstructive(bool obstructive)
{
  if (CurrentDoodad != -1)
    Doodads[CurrentDoodad].doodad->obstructive = obstructive;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ClearParty()
{
  ME_ResetParty();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::AddPartyCharacter(const char* spriteset)
{
  // add character
  char path[520];
  sprintf(path, "spritesets/%s", spriteset);
  FILE* file = m_FileServer.GetFile(path);

  SSPRITESET* ss = m_SpritesetServer.Load(spriteset, file);
  if (ss == NULL)
  {
    char exitmessage[520];
    sprintf(exitmessage, "Error: Could not load '%s'", spriteset);
    ExitWithMessage(exitmessage);
    return;
  }
  
  int size = Party.characters.size();
  Party.characters.resize(size + 1);
  Party.characters[size].direction  = Party.location.direction;
  Party.characters[size].mapx       = Party.location.mapx;
  Party.characters[size].mapy       = Party.location.mapy;
  Party.characters[size].maplayer   = Party.location.maplayer;
  Party.characters[size].nextswitch = 0;
  Party.characters[size].walkframe  = 0;
  Party.characters[size].spriteset  = ss;

  // add to train
  Party.train.resize((size + 1) * 16);

  for (int i = size * 16; i < (size + 1) * 16; i++)
  {
    if (i == 0)
    {
      Party.train[i].mapx      = Party.location.mapx;
      Party.train[i].mapy      = Party.location.mapy;
      Party.train[i].maplayer  = Party.location.maplayer;
      Party.train[i].direction = Party.location.direction;
    }
    else
      Party.train[i] = Party.train[i - 1];
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::SetPersonMode(const char* person, int mode)
{
  PERSON* p = ME_FindPerson(person);
  if (p) {
    // clear command queues
    while (p->commands.empty() == false) {
      p->commands.pop();
    }
    p->mode = mode;
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::QueueCommand(const char* person, int command, bool immediate)
{
  PERSON* p = ME_FindPerson(person);
  if (p) {
    PERSON::Command c;
    c.value = command;
    c.immediate = immediate;
    p->commands.push(c);
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ClearCommandQueue(const char* person)
{
  PERSON* p = ME_FindPerson(person);
  if (p) {
    // clear command queues
    while (p->commands.empty() == false) {
      p->commands.pop();
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::PlayAnimation(const char* filename)
{
  char path[1024];
  sprintf(path, "anim/%s", filename);
  FILE* file = m_FileServer.GetFile(path);

  if (!::PlayAnimation(file))
  {
    char message[1024];
    sprintf(message, "Could not load animation '%s'", filename);
    ExitWithMessage(message);
  }
}

////////////////////////////////////////////////////////////////////////////////

DATAFILE*
CGameEngine::OpenFile(const char* filename)
{
  DATAFILE* file = new DATAFILE;
  file->config = new CONFIG;
  file->modified = false;
  char path[520];

  sprintf(path, "save/%s", filename);
  file->filename = newstr(path);
  FILE* fp = m_FileServer.GetFile(path);
  long  len = m_FileServer.GetFileLength(path);

  LoadConfigFromFile(file->config, fp, len);
  return file;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::CloseFile(DATAFILE* file)
{
  MakeDirectory("save");
  SaveConfig(file->config, file->filename);
  DestroyConfig(file->config);
  delete file->config;
  delete[] file->filename;
  delete file;
}

////////////////////////////////////////////////////////////////////////////////

MENUDISPLAYDATA
CGameEngine::GetMenuDisplayData()
{
  MENUDISPLAYDATA mdd;
  mdd.arrow = Arrow;
  mdd.up_arrow = UpArrow;
  mdd.down_arrow = DownArrow;
  mdd.font = &Font;
  mdd.window_style = &WindowStyle;
  return mdd;
}

////////////////////////////////////////////////////////////////////////////////

bool
CGameEngine::InitializeGame()
{
  CurrentMask.red   = 0;
  CurrentMask.green = 0;
  CurrentMask.blue  = 0;
  CurrentMask.alpha = 0;

  TargetMask.red    = 0;
  TargetMask.green  = 0;
  TargetMask.blue   = 0;
  TargetMask.alpha  = 0;

  // load the game.inf file
  char scriptname[520];
  ReadConfigFileString("game.inf", "", "script", scriptname, 512, "");

  // get the game resolution and set it
  int screen_width, screen_height;
  ReadConfigFileInt("game.inf", "", "screen_width", &screen_width, 320);
  ReadConfigFileInt("game.inf", "", "screen_height", &screen_height, 240);
  SwitchResolution(screen_width, screen_height);

  // load the default font, window style, and arrows
  Font.Load(m_FileServer.GetFile(m_SystemFiles->font));
  WindowStyle.Load(m_FileServer.GetFile(m_SystemFiles->window_style));
  Arrow = SLoadImage(m_FileServer.GetFile(m_SystemFiles->arrow));
  UpArrow = SLoadImage(m_FileServer.GetFile(m_SystemFiles->up_arrow));
  DownArrow = SLoadImage(m_FileServer.GetFile(m_SystemFiles->down_arrow));

  ME_InitParty();

  // scripts are in the "scripts" directory
  char scriptpath[520];
  sprintf(scriptpath, "scripts/%s", scriptname);

  // allocate the game script
  GameScript = new CScript();
  if (!GameScript->Load(scriptpath, m_FileServer.GetFile(scriptpath)))
  {
    delete GameScript;
    GameScript = NULL;
    char ErrorMessage[520];
    sprintf(ErrorMessage, "Error: Could not load game script '%s'", scriptname);
    ExitWithMessage(ErrorMessage);
    return false;
  }

  return true;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ShutdownGame()
{
  // if arrow somehow got destroyed, and nobody loaded a new one...
  if (Arrow)
    DestroyImage(Arrow);

  ME_ResetParty();

  SetMap("");
  PlayMusic("", NULL, 0);

  // switch resolution back to normal
  SwitchResolution(GetDefaultWidth(), GetDefaultHeight());

  delete GameScript;
  GameScript = NULL;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::RunGame()
{
  if (GameScript->ExecuteFunction("game") == false)
    ExitWithMessage("Error: Game script did not contain the function 'game'");
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ShowErrorScreen()
{
  // reset the font to the system one
  Font.Load(m_FileServer.GetFile(m_SystemFiles->font));

  // If an error occured, show the error screen
  if (ShouldExitEngine && ExitEngineMessage.length() > 0)
  {
    WhileAnyKeyPressed();

    do
    {
      ClearScreen();
      Font.DrawTextBox(0, 0, GetScreenWidth(), GetScreenHeight(), 0, ExitEngineMessage.c_str());
      FlipScreen();
      RefreshInput();
    }
    while (!AnyKeyPressed());

    ShouldExitEngine = false;
  }

  // wait until a key is not pressed
  WhileAnyKeyPressed();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ProcessFrameHooks()
{
  for (int i = 0; i < FrameHooks.size(); i++)
    if (++FrameHooks[i].current >= FrameHooks[i].update_rate)
    {
      GameScript->ExecuteFunction(FrameHooks[i].function.c_str());
      FrameHooks[i].current = 0;
    }  
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ProcessPersons()
{
  for (int i = 0; i < Persons.size(); i++) {
    ME_ProcessPerson(Persons[i]);
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ProcessPerson(PERSON& p)
{
  // if there are no commands, add to the command queue
  if (p.commands.size() == 0) {
    switch (p.mode) {
      case sPersonEntity::NONE: {  // he doesn't need any commands
        break;
      }

      case sPersonEntity::RANDOM: {

        PERSON::Command command;
        command.immediate = true;

        // add a command to change direction
        int direction = rand() % 4;
        switch (direction) {
          case 0: command.value = PERSON::Command::FACE_NORTH; break;
          case 1: command.value = PERSON::Command::FACE_EAST;  break;
          case 2: command.value = PERSON::Command::FACE_SOUTH; break;
          case 3: command.value = PERSON::Command::FACE_WEST;  break;
        }
        p.commands.push(command);

        command.immediate = false;
        
        switch (direction) {
          case 0: command.value = PERSON::Command::WALK_NORTH; break;
          case 1: command.value = PERSON::Command::WALK_EAST;  break;
          case 2: command.value = PERSON::Command::WALK_SOUTH; break;
          case 3: command.value = PERSON::Command::WALK_WEST;  break;
        }

        // add some walking commands
        for (int i = 0; i < Map.GetTileset().GetTileWidth(); i++) {
          p.commands.push(command);
        }

        // add a second of waiting
        command.value = PERSON::Command::WAIT;
        for (int i = 0; i < FPS; i++) {
          p.commands.push(command);
        }
        
        break;
      }

      case sPersonEntity::REPEATING: {
        // nothing to fill the command queue with (it should already have stuff)!
        break;
      }

    } // end switch
    return;
  }

  // process commands until one isn't immediate
  bool done = false;
  while (!done && p.commands.size()) {
    PERSON::Command c = p.commands.front();
    p.commands.pop();

    // if the person is repeating, add it to the end of the command queue
    if (p.mode == sPersonEntity::REPEATING) {
      p.commands.push(c);
    }

    // store old position to see if it changed
    int old_x = p.person->x;
    int old_y = p.person->y;

    switch (c.value) {
      case PERSON::Command::WALK_NORTH:     p.person->y--;                   break;
      case PERSON::Command::WALK_NORTHEAST: p.person->y--; p.person->x++;    break;
      case PERSON::Command::WALK_EAST:      p.person->x++;                   break;
      case PERSON::Command::WALK_SOUTHEAST: p.person->x++; p.person->y++;    break;
      case PERSON::Command::WALK_SOUTH:     p.person->y++;                   break;
      case PERSON::Command::WALK_SOUTHWEST: p.person->y++; p.person->x--;    break;
      case PERSON::Command::WALK_WEST:      p.person->x--;                   break;
      case PERSON::Command::WALK_NORTHWEST: p.person->x--; p.person->y--;    break;
      case PERSON::Command::FACE_NORTH:     p.direction = PERSON::NORTH;     break;
      case PERSON::Command::FACE_NORTHEAST: p.direction = PERSON::NORTHEAST; break;
      case PERSON::Command::FACE_EAST:      p.direction = PERSON::EAST;      break;
      case PERSON::Command::FACE_SOUTHEAST: p.direction = PERSON::SOUTHEAST; break;
      case PERSON::Command::FACE_SOUTH:     p.direction = PERSON::SOUTH;     break;
      case PERSON::Command::FACE_SOUTHWEST: p.direction = PERSON::SOUTHWEST; break;
      case PERSON::Command::FACE_WEST:      p.direction = PERSON::WEST;      break;    
      case PERSON::Command::FACE_NORTHWEST: p.direction = PERSON::NORTHWEST; break;
    }

    // if position changed, update walkframe
    if (old_x != p.person->x || old_y != p.person->y) {
      p.frame++;
    }

    done = !c.immediate;
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ProcessInput()
{
  RefreshInput();

  if (KeyPressed(KEY_F1) && !FPSKeyPressed)
    ShowFPS = !ShowFPS;
  
  FPSKeyPressed = KeyPressed(KEY_F1);

  // Store backups of the current coordinates
  int oldmapx = Party.location.mapx;
  int oldmapy = Party.location.mapy;

  if (KeyPressed(KEY_LEFT))
    Party.location.mapx--;
  if (KeyPressed(KEY_RIGHT))
    Party.location.mapx++;
  if (KeyPressed(KEY_UP))
    Party.location.mapy--;
  if (KeyPressed(KEY_DOWN))
    Party.location.mapy++;
  if (KeyPressed(KEY_ESCAPE))
    ShouldExitEngine = true;

  int& mx = Party.location.mapx;
  int& my = Party.location.mapy;
  int& ox = oldmapx;
  int& oy = oldmapy;

  // update directions
  if (mx > ox) // East
  {
    if (my > oy)  Party.location.direction = 3; else // Southeast
    if (my == oy) Party.location.direction = 2; else // East
    if (my < oy)  Party.location.direction = 1;      // Northeast
  }
  else if (mx < ox) // West
  {
    if (my > oy)  Party.location.direction = 5; else // Southwest
    if (my == oy) Party.location.direction = 6; else // West
    if (my < oy)  Party.location.direction = 7;      // Northwest
  }
  else
  {
    if (my > oy) Party.location.direction = 4; else // South
    if (my < oy) Party.location.direction = 0;      // North
  }

  // make the first guy in the party face in the direction of the whole party
  if (Party.characters.size() > 0)
    Party.characters[0].direction = Party.location.direction;

  // obstruction checks
  if (Party.characters.size() > 0)  // people have to be in the party to have obstructions
  {
    const sSpriteset& spriteset = Party.characters[0].spriteset->GetSpriteset();
    int layer = Party.location.maplayer;

    // if the spriteset is currently obstructed, we must find a way to move it back
    if (ME_Obstructed(spriteset, mx, my, layer))
    {
      // motion is horizontal
      if (mx != ox && my == oy)
      {
        if (!ME_Obstructed(spriteset, mx, my - 1, layer))
          my--;
        else if (!ME_Obstructed(spriteset, mx, my + 1, layer))
          my++;
        else
        {
          mx = ox;
          my = oy;
        }
      }
      // motion is vertical
      else if (mx == ox && my != oy)
      {
        if (!ME_Obstructed(spriteset, mx - 1, my, layer))
          mx--;
        else if (!ME_Obstructed(spriteset, mx + 1, my, layer))
          mx++;
        else
        {
          mx = ox;
          my = oy;
        }
      }
      // motion is diagonal
      else if (mx != ox && my != oy)
      {
        if (!ME_Obstructed(spriteset, mx, oy, layer))
          my = oy;
        else if (!ME_Obstructed(spriteset, ox, my, layer))
          mx = ox;
        else
        {
          mx = ox;
          my = oy;
        }
      }
    }
  }

  // update train and walkframes
  if (mx != ox || my != oy)
  {
    // update train
    for (int i = Party.characters.size() * 16 - 1; i >= 0; i--)
    {
      if (i == 0)
      {
        Party.train[i].mapx      = Party.location.mapx;
        Party.train[i].mapy      = Party.location.mapy;
        Party.train[i].maplayer  = Party.location.maplayer;
        Party.train[i].direction = Party.location.direction;
      }
      else
        Party.train[i] = Party.train[i - 1];

      if (i % 16 == 0)
      {
        Party.characters[i / 16].mapx      = Party.train[i].mapx;
        Party.characters[i / 16].mapy      = Party.train[i].mapy;
        Party.characters[i / 16].maplayer  = Party.train[i].maplayer;
        Party.characters[i / 16].direction = Party.train[i].direction;
      }
    }

    // update walkframes
    for (int i = 0; i < Party.characters.size(); i++)
    {
      PARTY::PARTYCHARACTER* character = &Party.characters[i];
      character->nextswitch--;
      if (character->nextswitch < 0)
      {
        character->walkframe = (character->walkframe + 1) % character->spriteset->GetNumFrames(character->direction);

        character->nextswitch = character->spriteset->GetFrameDelay(
          character->direction,
          character->walkframe);
      }
    }
  }

  if (CameraTracking)
  {
    Camera.x = Party.location.mapx;
    Camera.y = Party.location.mapy;
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ProcessBoundKeys()
{
  std::map<int, std::string>::iterator i = BoundKeys.begin();
  while (i != BoundKeys.end())
  {
    if (KeyPressed((*i).first))
      if (GameScript)
        GameScript->ExecuteFunction((*i).second.c_str());

    i++;
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_CheckWarps()
{
  const sMap& map = Map.GetMap();
  for (int i = 0; i < map.GetNumEntities(); i++)
  {
    if (map.GetEntity(i).GetEntityType() != sEntity::WARP)
      continue;

    int tile_width = map.GetTileset().GetTileWidth();
    int tile_height = map.GetTileset().GetTileHeight();

    const sWarpEntity& warp = (const sWarpEntity&)map.GetEntity(i);
    if (Party.location.mapx / tile_width  == warp.x / tile_width &&
        Party.location.mapy / tile_height == warp.y / tile_height &&
        Party.location.maplayer == warp.layer)
    {
      ME_RenderScreen();
      FadeOut(warp.fade);

      SetMap(warp.destination_map.c_str());

      ME_RenderScreen();
      FadeIn(warp.fade);
    }
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_CheckTriggers()
{
  bool on_trigger = false;

  const sMap& map = Map.GetMap();
  for (int i = 0; i < map.GetNumEntities(); i++)
  {
    if (map.GetEntity(i).GetEntityType() != sEntity::TRIGGER)
      continue;

    int tile_width = map.GetTileset().GetTileWidth();
    int tile_height = map.GetTileset().GetTileHeight();

    const sTriggerEntity& trigger = (const sTriggerEntity&)map.GetEntity(i);
    if (Party.location.mapx / tile_width == trigger.x / tile_width &&
        Party.location.mapy / tile_height == trigger.y / tile_height &&
        Party.location.maplayer == trigger.layer)
    {
      if (!OnTrigger)
      {
        if (MapScript)
          MapScript->ExecuteFunction(trigger.function.c_str());
      }

      on_trigger = true;
    }
  }

  OnTrigger = on_trigger;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_CheckDoodads()
{
  // check DOOR activation
  for (int i = 0; i < Doodads.size(); i++)
    if (Doodads[i].doodad->activation_method == sDoodadEntity::DOOR)
    {
      int ox = 1;
      int oy = 1;

      // ox and oy should be (the size of the base rectangle) / 2
      if (Party.characters.size() > 0)
      {
        const sSpriteset& ss = Party.characters[0].spriteset->GetSpriteset();
        int width = abs(ss.GetBaseX1() - ss.GetBaseX2());
        int height = abs(ss.GetBaseY1() - ss.GetBaseY2());
        ox = width / 2 + 2;  // I don't know why there needs to be a +2 on here
        oy = height / 2 + 2;
      }

      // check obstructions
      int mapx = Party.location.mapx;
      int mapy = Party.location.mapy;
      if (ME_ObstructedDoodad(i, mapx + ox, mapy + 0)  ||
          ME_ObstructedDoodad(i, mapx + ox, mapy + oy) ||
          ME_ObstructedDoodad(i, mapx + 0,  mapy + oy) ||
          ME_ObstructedDoodad(i, mapx - ox, mapy + oy) ||
          ME_ObstructedDoodad(i, mapx - ox, mapy + 0)  ||
          ME_ObstructedDoodad(i, mapx - ox, mapy - oy) ||
          ME_ObstructedDoodad(i, mapx + 0,  mapy - oy))
      {
        // activate the doodad
        CurrentDoodad = i;
        if (MapScript)
          MapScript->ExecuteFunction(Doodads[i].doodad->function.c_str());
        CurrentDoodad = -1;
      }
    }
}

////////////////////////////////////////////////////////////////////////////////

inline int min(int a, int b)
{
  return (a < b ? a : b);
}

inline int max(int a, int b)
{
  return (a > b ? a : b);
}

bool
CGameEngine::ME_Obstructed(const sSpriteset& ss, int mapx, int mapy, int maplayer)
{
  const sTileset& ts = Map.GetTileset();
  
  int base_min_x = min(ss.GetBaseX1(), ss.GetBaseX2());
  int base_max_x = max(ss.GetBaseX1(), ss.GetBaseX2());
  int base_min_y = min(ss.GetBaseY1(), ss.GetBaseY2());
  int base_max_y = max(ss.GetBaseY1(), ss.GetBaseY2());

  int hotspot_x = (base_max_x + base_min_x) / 2;
  int hotspot_y = (base_max_y + base_min_y) / 2;

  for (int ix = base_min_x; ix <= base_max_x; ix++)
    for (int iy = base_min_y; iy <= base_max_y; iy++)
    {
      int map_x = mapx + ix - hotspot_x;
      int map_y = mapy + iy - hotspot_y;

      int tx = map_x / ts.GetTileWidth();
      int ty = map_y / ts.GetTileHeight();
      const sLayer& layer = Map.GetMap().GetLayer(maplayer);
      if (tx > 0 && ty > 0 && tx < layer.GetWidth() && ty < layer.GetHeight())
      {
        const sTile& tile = ts.GetTile(layer.GetTile(tx, ty));

        int tile_offset_x = map_x % ts.GetTileWidth();
        int tile_offset_y = map_y % ts.GetTileHeight();

        if (tile.IsObstructed(tile_offset_x, tile_offset_y))
          return true;
      }

      // check if the area is obstructed by doodads
      if (ME_ObstructedDoodads(map_x, map_y))
        return true;
    }

  return false;
}

////////////////////////////////////////////////////////////////////////////////

bool
CGameEngine::ME_ObstructedDoodad(int doodad, int map_x, int map_y)
{
  // now check if map_x and map_y are between the spriteset's base
  const sSpriteset& ss = Doodads[doodad].spriteset->GetSpriteset();
  int width  = abs(ss.GetBaseX1() - ss.GetBaseX2());
  int height = abs(ss.GetBaseY1() - ss.GetBaseY2());

  int x = Doodads[doodad].doodad->x;
  int y = Doodads[doodad].doodad->y;

  return (
    map_x >= x - width / 2 &&
    map_x <= x + width / 2 &&
    map_y >= y - height / 2 &&
    map_y <= y + height / 2
  );
}

////////////////////////////////////////////////////////////////////////////////

bool
CGameEngine::ME_ObstructedDoodads(int map_x, int map_y)
{
  for (int i = 0; i < Doodads.size(); i++)
    if (Doodads[i].doodad->obstructive)
      if (ME_ObstructedDoodad(i, map_x, map_y))
        return true;
  return false;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_InitColorMask()
{
  PreviousMask.red   = 0;
  PreviousMask.green = 0;
  PreviousMask.blue  = 0;
  PreviousMask.alpha = 0;

  CurrentMask.red   = 0;
  CurrentMask.green = 0;
  CurrentMask.blue  = 0;
  CurrentMask.alpha = 0;

  TargetMask.red   = 0;
  TargetMask.green = 0;
  TargetMask.blue  = 0;
  TargetMask.alpha = 0;

  FadeTime        = 0;
  CurrentFadeTime = 0;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_UpdateColorMask()
{
  // if FadeTime is 0, we're not updating the mask
  if (FadeTime == 0)
    return;

  CurrentMask.red   = (PreviousMask.red   * (FadeTime - CurrentFadeTime)) + (TargetMask.red   * CurrentFadeTime) / FadeTime;
  CurrentMask.green = (PreviousMask.green * (FadeTime - CurrentFadeTime)) + (TargetMask.green * CurrentFadeTime) / FadeTime;
  CurrentMask.blue  = (PreviousMask.blue  * (FadeTime - CurrentFadeTime)) + (TargetMask.blue  * CurrentFadeTime) / FadeTime;
  CurrentMask.alpha = (PreviousMask.alpha * (FadeTime - CurrentFadeTime)) + (TargetMask.alpha * CurrentFadeTime) / FadeTime;

  CurrentFadeTime++;
  if (CurrentFadeTime > FadeTime - 1)
  {
    PreviousMask = TargetMask;
    CurrentMask = TargetMask;
    FadeTime = 0;
    CurrentFadeTime = 0;
  }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_RenderScreen()
{
  ClearScreen();

  for (int i = 0; i < Map.GetMap().GetNumLayers(); i++)
    ME_DrawLayer(i);
  
  ApplyColorMask(CurrentMask);
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_DrawLayer(int layer)
{
  ME_PreDrawObjects(layer);

  // only draw layer if it's visible
  if (Map.GetMap().GetLayer(layer).IsVisible())
  {
    // get center of screen
    int cx, cy;
    ME_GetCenterCoordinates(&cx, &cy);

    int tile_width  = Map.GetTileset().GetTileWidth();
    int tile_height = Map.GetTileset().GetTileHeight();

    // get tile coordinates and offset coordinates
    int tx = cx / tile_width;
    int ty = cy / tile_height;
    int ox = cx % tile_width;
    int oy = cy % tile_height;

    int half_width = GetScreenWidth() / 2;
    int half_height = GetScreenHeight() / 2;

    // draw layer!
    for (int ix = -half_width / tile_width - 2; ix < half_width / tile_width + 2; ix++)
      for (int iy = -half_height / tile_height - 2; iy < half_height / tile_height + 2; iy++)
      {
        if ((tx + ix) >= 0 &&
            (ty + iy) >= 0 &&
            (tx + ix) < Map.GetMap().GetLayer(layer).GetWidth() &&
            (ty + iy) < Map.GetMap().GetLayer(layer).GetHeight())
        {
          int tile = Map.GetMap().GetLayer(layer).GetTile(tx + ix, ty + iy);
          BlitImage(Map.GetTileImage(tile),
                    ix * tile_width  + half_width  - ox,
                    iy * tile_height + half_height - oy);
        }
      }
  }

  ME_PostDrawObjects(layer);
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_PreDrawObjects(int layer)
{
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_PostDrawObjects(int layer)
{
  CRenderSort rs;

  // add party characters
  for (int i = 0; i < Party.characters.size(); i++)
  {
    PARTY::PARTYCHARACTER* c = &Party.characters[i];
    if (c->maplayer == layer)
    {
      // calculate the offset of the hotspot in the sprite
      const sSpriteset& spriteset = c->spriteset->GetSpriteset();
      int offset_x = abs(spriteset.GetBaseX1() + spriteset.GetBaseX2()) / 2;
      int offset_y = abs(spriteset.GetBaseY1() + spriteset.GetBaseY2()) / 2;

      // calculate the screen location of the object
      int base_x, base_y;
      MapToScreen(c->mapx, c->mapy, &base_x, &base_y);

      IMAGE image = c->spriteset->GetFrame(c->direction, c->walkframe);
      rs.AddObject(base_x, base_y, offset_x, offset_y, image);
    }
  }

  // add persons
  for (int i = 0; i < Persons.size(); i++)
    if (Persons[i].person->layer == layer)
    {
      const sSpriteset& spriteset = Persons[i].spriteset->GetSpriteset();
      int offset_x = abs(spriteset.GetBaseX1() + spriteset.GetBaseX2()) / 2;
      int offset_y = abs(spriteset.GetBaseY1() + spriteset.GetBaseY2()) / 2;

      // calculate the screen location of the object
      int base_x, base_y;
      MapToScreen(Persons[i].person->x, Persons[i].person->y, &base_x, &base_y);

      int frame = (Persons[i].frame / 8) % Persons[i].spriteset->GetNumFrames(Persons[i].direction);
      IMAGE image = Persons[i].spriteset->GetFrame(Persons[i].direction, frame);
      rs.AddObject(base_x, base_y, offset_x, offset_y, image);
    }

  // add doodads
  for (int i = 0; i < Doodads.size(); i++)
    if (Doodads[i].doodad->layer == layer)
    {
      // calculate the offset of the hotspot in the sprite
      const sSpriteset& spriteset = Doodads[i].spriteset->GetSpriteset();
      int offset_x = abs(spriteset.GetBaseX1() + spriteset.GetBaseX2()) / 2;
      int offset_y = abs(spriteset.GetBaseY1() + spriteset.GetBaseY2()) / 2;

      // calculate the screen location of the object
      int base_x, base_y;
      MapToScreen(Doodads[i].doodad->x, Doodads[i].doodad->y, &base_x, &base_y);

      IMAGE image = Doodads[i].spriteset->GetFrame(0, Doodads[i].current_frame);
      rs.AddObject(base_x, base_y, offset_x, offset_y, image);
    }
  
  rs.DrawObjects();
}

////////////////////////////////////////////////////////////////////////////////

// In:  map coordinates (in pixels)
// Out: screen coordinates (in pixels)
void
CGameEngine::MapToScreen(int mapx, int mapy, int* screenx, int* screeny)
{
  *screenx = GetScreenWidth() / 2;
  *screeny = GetScreenHeight() / 2;

  int cx, cy;
  ME_GetCenterCoordinates(&cx, &cy);

  *screenx -= (cx - mapx);
  *screeny -= (cy - mapy);
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_InitParty()
{
  Party.location.mapx = 0;
  Party.location.mapy = 0;
  Party.location.maplayer = 0;
  Party.location.direction = 0;

  Party.characters.clear();
  Party.train.clear();
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_ResetParty()
{
  for (int i = 0; i < Party.characters.size(); i++)
    m_SpritesetServer.Free(Party.characters[i].spriteset);

  Party.characters.clear();
  Party.train.clear();

  Party.location.direction = 0;
  Party.location.mapx      = 0;
  Party.location.mapy      = 0;
  Party.location.maplayer  = 0;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_AddPersons()
{
  sMap& map = Map.GetMap();
  for (int i = 0; i < map.GetNumEntities(); i++)
    if (map.GetEntity(i).GetEntityType() == sEntity::PERSON)
    {
      PERSON p;
      FILE *file;
      char path[520];
      p.person = (sPersonEntity*)&map.GetEntity(i);
      sprintf(path, "spritesets/%s", p.person->spriteset.c_str());
      file = m_FileServer.GetFile(path);

      p.spriteset = m_SpritesetServer.Load(p.person->spriteset.c_str(), file);
      if (p.spriteset == NULL)
      {
        for (int j = 0; j < Persons.size(); j++)
          m_SpritesetServer.Free(Persons[j].spriteset);
        Persons.clear();

        char message[FILENAME_MAX + 100];
        sprintf(message, "Error: Cannot load spriteset '%s'", p.person->spriteset.c_str());
        ExitWithMessage(message);
        return;
      }

      p.mode        = p.person->movement_type;
      p.direction   = PERSON::SOUTH;
      p.frame       = 0;
      Persons.push_back(p);
    }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_RemovePersons()
{
  for (int j = 0; j < Persons.size(); j++)
    m_SpritesetServer.Free(Persons[j].spriteset);
  Persons.clear();
}

////////////////////////////////////////////////////////////////////////////////

CGameEngine::PERSON*
CGameEngine::ME_FindPerson(const char* name)
{
  for (int i = 0; i < Persons.size(); i++) {
    if (name == Persons[i].person->name) {
      return &Persons[i];
    }
  }
  return NULL;
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_AddDoodads()
{
  sMap& map = Map.GetMap();
  for (int i = 0; i < map.GetNumEntities(); i++)
    if (map.GetEntity(i).GetEntityType() == sEntity::DOODAD)
    {
      DOODAD d;
      FILE* file;
      char path[520];
      d.current_frame = 0;
      d.doodad = (sDoodadEntity*)&map.GetEntity(i);
      sprintf(path, "spritesets/%s", d.doodad->spriteset_file.c_str());
      file = m_FileServer.GetFile(path);

      d.spriteset = m_SpritesetServer.Load(d.doodad->spriteset_file.c_str(), file);
      if (d.spriteset == NULL)
      {
        for (int j = 0; j < Doodads.size(); j++)
          m_SpritesetServer.Free(Doodads[j].spriteset);
        Doodads.clear();

        char message[FILENAME_MAX + 100];
        sprintf(message, "Error: Cannot load spriteset '%s'", d.doodad->spriteset_file.c_str());
        ExitWithMessage(message);
        return;
      }
      Doodads.push_back(d);
    }
}

////////////////////////////////////////////////////////////////////////////////

void
CGameEngine::ME_RemoveDoodads()
{
  for (int i = 0; i < Doodads.size(); i++)
    m_SpritesetServer.Free(Doodads[i].spriteset);
  Doodads.clear();
}

////////////////////////////////////////////////////////////////////////////////

// Returns the map coordinates (in pixels, not tiles) that are the center of the screen
void
CGameEngine::ME_GetCenterCoordinates(int* cx, int* cy)
{
  *cx = Camera.x;
  *cy = Camera.y;

  int tile_width  = Map.GetTileset().GetTileWidth();
  int tile_height = Map.GetTileset().GetTileHeight();
  int half_width  = GetScreenWidth() / 2;
  int half_height = GetScreenHeight() / 2;
  if (*cx < half_width)
    *cx = half_width;
  if (*cy < half_height)
    *cy = half_height;

  if (*cx > Map.GetMap().GetLayer(Party.location.maplayer).GetWidth() * tile_width - half_width)
    *cx = Map.GetMap().GetLayer(Party.location.maplayer).GetWidth() * tile_width - half_width;

  if (*cy > Map.GetMap().GetLayer(Party.location.maplayer).GetHeight() * tile_height - half_height)
    *cy = Map.GetMap().GetLayer(Party.location.maplayer).GetHeight() * tile_height - half_height;
}

////////////////////////////////////////////////////////////////////////////////
