#ifndef ENGINE_HPP
#define ENGINE_HPP


// identifier too long
#ifdef _MSC_VER
#pragma warning(disable : 4786)
#endif


#include <vector>
#include <queue>
#include <map>
#include <string>
#include "engineinterface.hpp"
#include "colorspace.h"
#include "sphere.hpp"
#include "smap.hpp"
#include "sspriteset.hpp"
#include "sfont.hpp"
#include "swindowstyle.hpp"
#include "script.hpp"



class CGameEngine : private IEngineInterface
{
public:
  CGameEngine(const SSystemFiles* system_files);
  ~CGameEngine();

  void Run(const char* game);

private:

  // implementation of VM interface

  virtual void Exit();
  virtual void ExitWithMessage(const char* message);
  virtual bool ShouldExit();

  virtual IMAGE LoadImage(const char* filename);
  virtual CImage32* LoadSurface(const char* filename);

  virtual void SetFont(const char* filename);
  virtual SFONT* GetFont();

  virtual void SetWindowStyle(const char* filename);
  virtual SWINDOWSTYLE* GetWindowStyle();

  virtual void AddFrameHook(int update_rate, const char* function);
  virtual void RemoveFrameHook(const char* function);
  virtual void MapEngine();
  virtual void UpdateMap();
  virtual void RenderMap();
  virtual void SetMap(const char* filename);
  virtual SMAP* GetMap();

  virtual void BindKey(int key, const char* function);
  virtual void UnbindKey(int key);
  
  virtual void SetColorMask(RGBA mask, int num_frames);

  virtual void SetDoodadFrame(int frame);
  virtual void SetDoodadObstructive(bool obstructive);

  virtual void ClearParty();
  virtual void AddPartyCharacter(const char* filename);

  virtual void SetPersonMode(const char* person, int mode);
  virtual void QueueCommand(const char* person, int command, bool immediate);
  virtual void ClearCommandQueue(const char* person);

  virtual void PlayAnimation(const char* filename);

  virtual DATAFILE* OpenFile(const char* filename);
  virtual void CloseFile(DATAFILE* file);

  virtual MENUDISPLAYDATA GetMenuDisplayData();

private:
  struct FrameHook
  {
    string function;
    int update_rate;
    int current;
  };

  struct CAMERA
  {
    // pixel coordinates
    int x;
    int y;
  };
  
  struct PARTY
  {
    struct PARTYCHARACTER
    {
      int direction;  // 0 = north, 1 = northeast, 2 = east, etc.
      int mapx;       // pixel coordinates
      int mapy;       // ..
      int maplayer;   // z coordinate

      int nextswitch;
      int walkframe;
      SSPRITESET* spriteset;
    };

    struct PARTYCOORDINATES
    {
      int direction;  // 0 = north, 1 = northeast, 2 = east, etc.
      int mapx;       // pixel coordinates
      int mapy;       // ..
      int maplayer;   // z coordinate
    };

    std::vector<PARTYCHARACTER> characters;

    PARTYCOORDINATES location;
    std::vector<PARTYCOORDINATES> train;
  };

  struct PERSON
  {
    enum Direction {
      NORTH     = 0,
      NORTHEAST = 1,
      EAST      = 2,
      SOUTHEAST = 3,
      SOUTH     = 4,
      SOUTHWEST = 5,
      WEST      = 6,
      NORTHWEST = 7,
    };

    struct Command {
      enum {
        WAIT           = 0,
        WALK_NORTH     = 1,
        WALK_NORTHEAST = 2,
        WALK_EAST      = 3,
        WALK_SOUTHEAST = 4,
        WALK_SOUTH     = 5,
        WALK_SOUTHWEST = 6,
        WALK_WEST      = 7,
        WALK_NORTHWEST = 8,
        FACE_NORTH     = 9,
        FACE_NORTHEAST = 10,
        FACE_EAST      = 11,
        FACE_SOUTHEAST = 12,
        FACE_SOUTH     = 13,
        FACE_SOUTHWEST = 14,
        FACE_WEST      = 15,
        FACE_NORTHWEST = 16,
      };

      bool immediate;
      int  value;
    };

    sPersonEntity* person;
    SSPRITESET*    spriteset;
    int            mode;
    Direction      direction;
    int            frame;

    std::queue<Command> commands;
  };

  struct DOODAD
  {
    sDoodadEntity* doodad;
    SSPRITESET*    spriteset;
    int            current_frame;
  };

private:
  bool InitializeGame();
  void ShutdownGame();
  void RunGame();
  void ShowErrorScreen();

  void ME_ProcessFrameHooks();
  void ME_ProcessPersons();
  void ME_ProcessPerson(PERSON& p);
  void ME_ProcessInput();
  void ME_ProcessBoundKeys();
  void ME_CheckWarps();
  void ME_CheckTriggers();
  void ME_CheckDoodads();
  bool ME_Obstructed(const sSpriteset& ss, int mapx, int mapy, int maplayer);
  bool ME_ObstructedDoodad(int doodad, int map_x, int map_y);
  bool ME_ObstructedDoodads(int map_x, int map_y);

  void ME_InitColorMask();
  void ME_UpdateColorMask();

  void ME_RenderScreen();
  void ME_DrawLayer(int layer);
  void ME_PreDrawObjects(int layer);
  void ME_PostDrawObjects(int layer);
  void MapToScreen(int mapx, int mapy, int* screenx, int* screeny);

  void ME_InitParty();
  void ME_ResetParty();

  void ME_AddPersons();
  void ME_RemovePersons();
  PERSON* ME_FindPerson(const char* name);
  
  void ME_AddDoodads();
  void ME_RemoveDoodads();

  void ME_GetCenterCoordinates(int* cx, int* cy);

private:
  const SSystemFiles* m_SystemFiles;

  IMAGE Arrow;
  IMAGE UpArrow;
  IMAGE DownArrow;

  SFONT        Font;
  SWINDOWSTYLE WindowStyle;

  bool ShouldExitEngine;
  std::string ExitEngineMessage;

  CScript* GameScript;

  // rendering flags
  bool ShowFPS;
  bool FPSKeyPressed;

  // map data
  bool     MapLoaded;
  SMAP     Map;
  CScript* MapScript;

  bool OnTrigger;

  // color mask variables
  RGBA PreviousMask;
  RGBA CurrentMask;
  RGBA TargetMask;
  int  FadeTime;        // stored in frames
  int  CurrentFadeTime; // stored in frames, must be >= 0 and < FadeTime

  std::vector<FrameHook> FrameHooks;
  std::map<int, std::string> BoundKeys;

  // camera
  CAMERA Camera;
  bool CameraTracking;

  PARTY Party;
  std::vector<PERSON> Persons;
  std::vector<DOODAD> Doodads;
  int CurrentDoodad;
};


#endif
